'use strict';

var _slicedToArray = function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"]) _i["return"](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError("Invalid attempt to destructure non-iterable instance"); } }; }();

var _Scope = require('./Scope');

var _Scope2 = _interopRequireDefault(_Scope);

var _fetch = require('./fetch');

var _fetch2 = _interopRequireDefault(_fetch);

var _debug = require('../debug');

var _debug2 = _interopRequireDefault(_debug);

var _functions = require('./functions');

var functions = _interopRequireWildcard(_functions);

var _types = require('./types');

var types = _interopRequireWildcard(_types);

var _sort = require('./sort');

var _sort2 = _interopRequireDefault(_sort);

var _match = require('./match');

var _match2 = _interopRequireDefault(_match);

var _scopeTools = require('./scopeTools');

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) newObj[key] = obj[key]; } } newObj.default = obj; return newObj; } }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _asyncToGenerator(fn) { return function () { var gen = fn.apply(this, arguments); return new Promise(function (resolve, reject) { function step(key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { return Promise.resolve(value).then(function (value) { step("next", value); }, function (err) { step("throw", err); }); } } return step("next"); }); }; }

const DEFAULT_LIMIT = 100;

// Implementation of equals that do not respect object equality and eschews deep compare
function equals(lhs, rhs) {
  if (typeof lhs === 'object' || typeof rhs === 'object') {
    if (lhs && rhs && lhs.toJSON && rhs.toJSON) {
      return lhs.toJSON() == rhs.toJSON();
    }
    return false;
  }
  return lhs == rhs;
}

function lt(lhs, rhs) {
  if (typeof lhs === 'object' || typeof rhs === 'object') {
    if (lhs && rhs && lhs.toJSON && rhs.toJSON) {
      return lhs.toJSON() < rhs.toJSON();
    }
    return false;
  }
  return lhs < rhs;
}

function gt(lhs, rhs) {
  if (typeof lhs === 'object' || typeof rhs === 'object') {
    if (lhs && rhs && lhs.toJSON && rhs.toJSON) {
      return lhs.toJSON() > rhs.toJSON();
    }
    return false;
  }
  return lhs > rhs;
}

// Slices an input. If the input is an array of scopes: slices each scope individually. If the
// input is a scope with an array: slices the scope
function slice(input, start, end) {
  // if (Array.isArray(input)) {
  //   return input.map(scope => scope.slice(start, end))
  // }
  if (Array.isArray(input)) {
    return input.slice(start, end);
  }
  // Input is a single scope
  return input.slice(start, end);
}

function first(input, offset) {
  // if (Array.isArray(input)) {
  //   return input.map(scope => scope.first(offset))
  // }
  if (Array.isArray(input)) {
    return input[offset];
  }
  // Input is a single scope
  return input.first(offset);
}

function shallowFlatten(value) {
  if (!Array.isArray(value)) {
    return value;
  }
  const result = [];
  value.forEach(item => {
    if (Array.isArray(item)) {
      result.push(...item);
    } else {
      result.push(item);
    }
  });
  return result;
}

module.exports = class Executor {

  constructor(options) {
    const operations = options.operations,
          scope = options.scope,
          fetcher = options.fetcher;

    this.scope = scope || new _Scope2.default({
      parent: null,
      value: {}
    });
    this.fetcher = fetcher;
    this.operations = operations;
  }

  run() {
    var _this = this;

    return _asyncToGenerator(function* () {
      return _this.exec(_this.operations, _this.scope);
    })();
  }

  exec(operation, scope) {
    var _this2 = this;

    return _asyncToGenerator(function* () {
      (0, _debug2.default)('exec() op:', operation, 'scope:', scope);
      if (scope instanceof Promise) {
        throw new Error('scopes should be plain values, not promises');
      }
      switch (operation.op) {
        case 'pipe':
          return _this2.execPipeline(operation, scope);
        case 'source':
          const data = scope.dataForSource(operation.id);
          return scope.child({
            sourceId: operation.id,
            value: data.documents,
            start: data.start
          });
        case 'filter':
          return _this2.execFilter(operation, scope);
        case 'and':
          return _this2.execAnd(operation, scope);
        case 'or':
          return _this2.execOr(operation, scope);
        case 'eq':
          return _this2.execEq(operation, scope);
        case 'neq':
          return _this2.execNEq(operation, scope);
        case 'lt':
          return _this2.execLT(operation, scope);
        case 'lte':
          return _this2.execLTE(operation, scope);
        case 'gt':
          return _this2.execGT(operation, scope);
        case 'gte':
          return _this2.execGTE(operation, scope);
        case 'not':
          return _this2.execNOT(operation, scope);
        case 'accessor':
          return _this2.execAccessor(operation, scope);
        case 'literal':
          return _this2.execLiteral(operation, scope);
        case 'object':
          return _this2.execObjectExpr(operation, scope);
        case 'array':
          return _this2.execArray(operation, scope);
        case 'in':
          return _this2.execIn(operation, scope);
        case 'functionCall':
          return _this2.execFunctionCall(operation, scope);
        case 'match':
          return _this2.execMatch(operation, scope);
        default:
          throw new Error(`(exec) Unknown operation ${operation.op}`);
      }
    })();
  }

  execFunctionCall(operation, scope) {
    var _this3 = this;

    return _asyncToGenerator(function* () {
      const fn = functions[operation.name];
      if (!fn) {
        throw new Error(`Unknown function ${operation.name}`);
      }
      const scopedArgs = yield Promise.all(operation.arguments.map(function (arg) {
        return _this3.exec(arg, scope);
      }));
      return scope.clone({
        value: fn(scope, ...scopedArgs)
        // value: fn(...(scopedArgs.map(scopedArg => scopedArg.value)))
      });
    })();
  }

  // Exec an array of operations and return their plain values as an array
  execEval(operations, scope) {
    var _this4 = this;

    return _asyncToGenerator(function* () {
      return Promise.all(operations.map(function (op) {
        return _this4.exec(op, scope);
      })).then(function (scopes) {
        return scopes.map(function (s) {
          return s.value;
        });
      });
    })();
  }

  execMatch(operation, scope) {
    var _this5 = this;

    return _asyncToGenerator(function* () {
      var _ref = yield Promise.all([_this5.exec(operation.lhs, scope), _this5.exec(operation.rhs, scope)]),
          _ref2 = _slicedToArray(_ref, 2);

      const lhs = _ref2[0],
            rhs = _ref2[1];

      const terms = Array.isArray(rhs.value) ? rhs.value : [rhs.value];
      return scope.clone({
        value: (0, _match2.default)(lhs.value, terms)
      });
    })();
  }

  execIn(operation, scope) {
    var _this6 = this;

    return _asyncToGenerator(function* () {
      var _ref3 = yield Promise.all([_this6.exec(operation.lhs, scope), _this6.exec(operation.rhs, scope)]),
          _ref4 = _slicedToArray(_ref3, 2);

      const lhs = _ref4[0],
            rhs = _ref4[1];

      const lhsValue = (0, _scopeTools.asPlainValue)(lhs);
      const rhsValue = (0, _scopeTools.asPlainValue)(rhs);
      if (lhsValue === null || lhsValue === undefined || rhsValue === null || rhsValue === undefined) {
        return scope.clone({
          value: null
        });
      }
      if (Array.isArray(rhsValue)) {
        for (let i = 0; i < rhsValue.length; i++) {
          const candidate = rhsValue[i];
          if (lhsValue == candidate) {
            return scope.clone({
              value: true
            });
          }
          if (candidate instanceof types.Path) {
            if (candidate.contains(lhsValue)) {
              return scope.clone({
                value: true
              });
            }
          }
        }
        return scope.clone({
          value: false
        });
      }
      if (rhsValue instanceof types.Path) {
        (0, _debug2.default)('path containment', lhsValue, rhsValue);
        return scope.clone({
          value: rhsValue.contains(lhsValue)
        });
      }
      throw new Error(`in-operator does not apply to rhs value ${rhs.value}`);
    })();
  }

  execArray(operation, scope) {
    var _this7 = this;

    return _asyncToGenerator(function* () {
      return Promise.all(operation.operations.map(function (op) {
        return _this7.exec(op, scope);
      })).then(function (elements) {
        return scope.clone({
          value: elements.map(function (element) {
            return (0, _scopeTools.asPlainValue)(element);
          })
        });
      });
    })();
  }

  execObjectExpr(operation, scope) {
    var _this8 = this;

    return _asyncToGenerator(function* () {
      const evalObjOp = (() => {
        var _ref5 = _asyncToGenerator(function* (objOp) {
          (0, _debug2.default)('objOp:', objOp);
          switch (objOp.op) {
            case 'assignment':
              return _this8.exec(objOp.value, scope).then(function (resultScope) {
                const value = (0, _scopeTools.asPlainValue)(resultScope);
                (0, _debug2.default)(`objOp assignment [${objOp.name}] = ${value}`, value);
                return {
                  [objOp.name]: value
                };
              });
            case 'splat':
              return scope.value;
            default:
              throw new Error(`Unknown object expression operation ${objOp.op}`);
          }
        });

        return function evalObjOp(_x) {
          return _ref5.apply(this, arguments);
        };
      })();

      return Promise.all(operation.operations.map(evalObjOp)).then(function (fragments) {
        (0, _debug2.default)('object fagments:', fragments);
        const value = Object.assign({}, ...fragments);
        const result = scope.clone({
          value
        });
        (0, _debug2.default)('projected object:', result);
        return result;
      });
    })();
  }

  pipeMap(operation, lhs) {
    var _this9 = this;

    return _asyncToGenerator(function* () {
      let input = Array.isArray(lhs) ? lhs : [lhs];
      return Promise.all(input.map(function (item) {
        return _this9.exec(operation, item);
      }));
    })();
  }

  pipeFilter(operation, lhs) {
    var _this10 = this;

    return _asyncToGenerator(function* () {
      let input;
      if (Array.isArray(lhs)) {
        (0, _debug2.default)("LHS is array");
        // If lhs is an array of scopes, we perform flattening on each scope before proceeding
        input = lhs;
        (0, _debug2.default)('flatten', lhs);
        input = lhs.reduce(function (sum, item) {
          if (Array.isArray(item.value)) {
            return sum.concat(item.value.map(function (element) {
              return item.clone({
                value: element
              });
            }));
          }
          return sum.concat(item);
        }, []);
        (0, _debug2.default)('flattened to', input);
      } else {
        // If lhs is a single array value, explode it to an array of scopes
        if (Array.isArray(lhs.value)) {
          (0, _debug2.default)("LHS is single array value (explode)");
          input = lhs.value.map(function (item) {
            return lhs.clone({
              value: item
            });
          });
        } else {
          // If lhs is a single non-array value, wrap it in an array
          (0, _debug2.default)("LHS is single non-array value (wrap)");
          input = [lhs];
        }
      }
      (0, _debug2.default)('mapped lhs for filter, lhs:', lhs, 'input:', input);
      const filterPromises = input.map(function (item) {
        return _this10.exec(operation, item);
      });
      return Promise.all(filterPromises).then(function (filterValues) {
        (0, _debug2.default)('filter result:', filterValues);
        return input.filter(function (element, index) {
          return filterValues[index].value === true;
        });
      });
    })();
  }

  pipe(operation, lhs) {
    var _this11 = this;

    return _asyncToGenerator(function* () {
      const inputIsArray = Array.isArray(lhs);
      let input = inputIsArray ? lhs : [lhs];
      let result;
      switch (operation.op) {
        case 'subscript':
          if (operation.first) {
            (0, _debug2.default)('first()', lhs, operation.start);
            (0, _debug2.default)('=>', first(lhs, operation.start));
            return first(lhs, operation.start);
          } else {
            (0, _debug2.default)('slice()', lhs, operation.start, operation.end);
            (0, _debug2.default)('=>', slice(lhs, operation.start, operation.end));
            return slice(lhs, operation.start, operation.end);
          }
          break;
        case 'ordering':
          return (0, _sort2.default)(lhs, operation.terms);
          break;
        case 'filter':
          return _this11.pipeFilter(operation, lhs);
          break;
        default:
          (0, _debug2.default)('mapping:', lhs);
          result = _this11.pipeMap(operation, lhs);
          break;
      }
      (0, _debug2.default)('input was array:', inputIsArray, result);
      if (inputIsArray) {
        return result;
      }
      if (!result.then) {
        (0, _debug2.default)("Un-arrayifying plain value", result);
        return Array.isArray(result) ? result[0] : result;
      }
      return result.then(function (arrayResult) {
        (0, _debug2.default)("Un-arrayifying promised value", arrayResult);
        return arrayResult[0];
      });
    })();
  }

  // NOTE: Consider keeping the pipeline value as an array of scope, not a scope
  // containing an array. Have a feeling each value in the mapping chain needs
  // the ability to keep track of it's own parents etc.
  execPipeline(operation, scope) {
    var _this12 = this;

    return _asyncToGenerator(function* () {
      scope = yield (0, _fetch2.default)(scope, operation, _this12.fetcher);
      const ops = operation.operations.slice();
      const first = ops.shift();
      let current = yield _this12.exec(first, scope);
      while (ops.length > 0) {
        (0, _debug2.default)('current:', current);
        let nextOp = ops.shift();
        (0, _debug2.default)('nextOp:', nextOp);
        current = yield _this12.pipe(nextOp, current);
      }
      (0, _debug2.default)('current:', current);
      return current;
    })();
  }

  execBinaryOp(operation, scope, impl) {
    var _this13 = this;

    return _asyncToGenerator(function* () {
      const lhs = yield _this13.exec(operation.lhs, scope);
      const rhs = yield _this13.exec(operation.rhs, scope);
      return new _Scope2.default({
        parent: scope.parent,
        value: impl(lhs.value, rhs.value)
      });
    })();
  }

  execPrefixOp(operation, scope, impl) {
    var _this14 = this;

    return _asyncToGenerator(function* () {
      const rhs = yield _this14.exec(operation.rhs, scope);
      return new _Scope2.default({
        parent: scope.parent,
        value: impl(rhs.value)
      });
    })();
  }

  execAnd(operation, scope) {
    var _this15 = this;

    return _asyncToGenerator(function* () {
      return _this15.execBinaryOp(operation, scope, function (lhs, rhs) {
        if (lhs === false || rhs === false) {
          return false;
        }
        if (typeof lhs !== 'boolean' || typeof rhs !== 'boolean') {
          return null;
        }
        return lhs && rhs;
      });
    })();
  }

  execOr(operation, scope) {
    var _this16 = this;

    return _asyncToGenerator(function* () {
      return _this16.execBinaryOp(operation, scope, function (lhs, rhs) {
        if (lhs === true || rhs === true) {
          return true;
        }
        if (typeof lhs !== 'boolean' || typeof rhs !== 'boolean') {
          return null;
        }
        return lhs || rhs;
      });
    })();
  }

  execEq(operation, scope) {
    var _this17 = this;

    return _asyncToGenerator(function* () {
      return _this17.execBinaryOp(operation, scope, function (lhs, rhs) {
        return equals(lhs, rhs);
      });
    })();
  }

  execNEq(operation, scope) {
    var _this18 = this;

    return _asyncToGenerator(function* () {
      return _this18.execBinaryOp(operation, scope, function (lhs, rhs) {
        return !equals(lhs, rhs);
      });
    })();
  }

  execLT(operation, scope) {
    var _this19 = this;

    return _asyncToGenerator(function* () {
      return _this19.execBinaryOp(operation, scope, function (lhs, rhs) {
        return lt(lhs, rhs);
      });
    })();
  }

  execLTE(operation, scope) {
    var _this20 = this;

    return _asyncToGenerator(function* () {
      return _this20.execBinaryOp(operation, scope, function (lhs, rhs) {
        return lt(lhs, rhs) || equals(lhs, rhs);
      });
    })();
  }

  execGT(operation, scope) {
    var _this21 = this;

    return _asyncToGenerator(function* () {
      return _this21.execBinaryOp(operation, scope, function (lhs, rhs) {
        return gt(lhs, rhs);
      });
    })();
  }

  execGTE(operation, scope) {
    var _this22 = this;

    return _asyncToGenerator(function* () {
      return _this22.execBinaryOp(operation, scope, function (lhs, rhs) {
        return gt(lhs, rhs) || equals(lhs, rhs);
      });
    })();
  }

  execNOT(operation, scope) {
    var _this23 = this;

    return _asyncToGenerator(function* () {
      return _this23.execPrefixOp(operation, scope, function (rhs) {
        if (typeof rhs !== 'boolean') {
          return null;
        }
        return !rhs;
      });
    })();
  }

  execAccessor(operation, scope) {
    return _asyncToGenerator(function* () {
      return scope.resolveAccessor(operation.path);
    })();
  }

  execLiteral(operation, scope) {
    return _asyncToGenerator(function* () {
      return new _Scope2.default({
        parent: null,
        value: operation.value
      });
    })();
  }

  execFilter(operation, scope) {
    var _this24 = this;

    return _asyncToGenerator(function* () {
      return _this24.exec(operation.filter, scope);
    })();
  }
};
//# sourceMappingURL=Executor.js.map