'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.path = path;
exports.joinPaths = joinPaths;
exports.coalesce = coalesce;
exports.count = count;
exports.length = length;
exports.defined = defined;
exports.references = references;

var _types = require('./types');

var types = _interopRequireWildcard(_types);

var _scopeTools = require('./scopeTools');

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) newObj[key] = obj[key]; } } newObj.default = obj; return newObj; } }

function path(_, spec) {
  return new types.Path(spec.value);
}

function joinPaths(_, a, b) {
  const aPath = new types.Path(a.value);
  const bPath = new types.Path(b.value);
  return aPath.concat(bPath);
}

function coalesce(_, ...args) {
  const result = args.find(arg => arg.value !== null && arg.value !== undefined);
  if (result === undefined) {
    return null;
  }
  return result.value;
}

function count(_, input) {
  const value = (0, _scopeTools.asPlainValue)(input);
  if (Array.isArray(value)) {
    return value.length;
  }
  return input.value !== null && input.value !== undefined ? 1 : 0;
}

function length(_, input) {
  const value = (0, _scopeTools.asPlainValue)(input);
  if (value === null || value === undefined) {
    return null;
  }
  if (Array.isArray(value)) {
    return value.length;
  }
  if (typeof value === 'string') {
    return value.length;
  }
  if (value && typeof value === 'object') {
    return Object.keys(value).length;
  }
  return value !== null && value !== undefined ? 1 : 0;
}

function defined(_, input) {
  if (Array.isArray(input.value)) {
    return input.value.length > 0;
  }
  if (input.value && typeof input.value === 'object') {
    return Object.keys(input.value).length > 0;
  }
  return input.value !== undefined && input.value !== null;
}

function references(scope, input) {
  const ids = input.value;
  return scope.doesReference(ids);
}
//# sourceMappingURL=functions.js.map